---------------------------------------------------------------
checkpoint()
---------------------------------------------------------------
NAME:
    checkpoint() - Logs the state of a C program.

SYNOPSIS:
    #include <stdlib.h>                 /* For EXIT_SUCCESS/FAILURE */
    #include "chkpoint.h"
    int checkpoint( void *save_area, unsigned sa_size, int last_flag );

DESCRIPTION:
    The checkpoint() function saves any information about files opened
    with ckp_open(), and any program information the programmer deems
    necessary to restart a program. Information recorded about open
    files is the pathname of the file, its open mode, and it current
    position.  To record program information the programmer must supply
    the address of a data object in memory, usually a structure, and the
    size of the data object.  All information is written to a checkpoint
    log file whose default directory is the current directory, but can be
    changed with the environment variable CHKPOINT_DIR.

    NOTE: Files opened as output mode are flushed to disk.

ARGUMENTS:
    void *save_area  - The address of a programmer supplied save area
                       to be written to disk.  Any information that
                       would be needed when a program is restarted
                       should be supplied with this argument. If this
                       argument is set to NULL no data will be saved.

    unsigned sa_size - This argument supplies the size of the
                       programmer supplied save area.  If this argument
                       is set to zero no data will be saved.

    int last_flag    - Normally this must be set to zero. When set to a
                       non-zero value checkpoint() goes through a
                       clean up process in which all files opened with
                       ckp_open() are flushed and closed and the
                       checkpoint log file is removed.  This process
                       should be reserved until the program's normal
                       processing is complete.

RETURN VALUE:
    EXIT_SUCCESS - The checkpoint was successful.
    EXIT_FAILURE - The checkpoint was not successful.  This is a
                   serious error.

SEE ALSO:
    ckp_open(), ckp_close(), restart().

---------------------------------------------------------------
ckp_close()
---------------------------------------------------------------
NAME:
    ckp_close() - Closes a file opened with ckp_open().

SYNOPSIS:
    #include <stdio.h>
    #include "chkpoint.h"
    int ckp_close( void *fp );

DESCRIPTION:
    This is a wrapper function for the standard fclose() function.  It
    removes a file from the list of files opened with ckp_open().  After
    ckp_close() has been used subsequent calls to checkpoint() do not
    log information about the file.

ARGUMENTS:
    void *fp - This is a standard file pointer returned from
               ckp_open() when a file is successfully opened.

RETURN VALUE:
    0 if successful.

SEE ALSO:
    ckp_open(), checkpoint(), restart().

---------------------------------------------------------------
ckp_open()
---------------------------------------------------------------
NAME:
    ckp_open() - Opens a file for use with checkpoint/restart.

SYNOPSIS:
    #include <stdio.h>
    #include "chkpoint.h"
    void *ckp_open( char *filepath, char *open_mode );

DESCRIPTION:
    The ckp_open() function is a wrapper for the standard fopen()
    function. This function is used to open the file, and add file
    information to a list of open files which will be used when the
    checkpoint() function is called.

    NOTE: When a program restarts, the restart() function re-opens all
    files that were open at the time of the last call to the
    checkpoint() function.  However, ckp_open() must be called
    regardless to obtain the file pointers of these files. Subsequent
    calls to ckp_open() with a filename which already open does not
    result in the file being opened a second time, the file's current
    file pointer is simply returned.

ARGUMENTS:
    char *filename  - Character string containing path and name of
                      the file to be opened.
    char *open_mode - String containing open mode information.
                      See fopen() for this information.

RETURN VALUE:
    A void pointer to an object of type FILE (in stdio.h).  If the
    pointer value is NULL there was an error opening the file.

SEE ALSO:
    ckp_close(), checkpoint(), restart().

---------------------------------------------------------------
restart()
---------------------------------------------------------------
NAME:
   restart() - Determines if a program is restarting.

SYNOPSIS:
    #include "chkpoint.h"
    int restart( char *ckp_id, void *save_area );

DESCRIPTION:
    The restart() function is used in a program in conjunction with the
    checkpoint() function to restart at program if the program
    abnormally terminated.  The restart() function should be called very
    early in program execution to determine if a program is restarting
    or not.  If a program is restarting the restart() function will open
    all files which were opened with the ckp_open() function at the time
    the last call to checkpoint() was made. In addition to re-opening
    the files restart() will also position the files at the point they
    were positioned when checkpoint() was last called.  Also, if any
    data was saved with checkpoint() restart() will restore it.

    NOTE:  The restart() function should be called prior to calling the
    checkpoint() function!  Also, restart() will use the value
    associated with the environment variable CHKPOINT_DIR to look for
    the checkpoint logfile.  If there is no value associated with
    CHKPOINT_DIR the current directory is used.  However, restart() will
    attempt to create a checkpoint logfile.  If it can it immediatly
    removes the logfile.  If it cannot create the file restart() returns
    an error condition.

ARGUMENTS:
    char *ckp_id    - A character string which should be unique for this
                      program.  This id is used as the major component of
                      the filename used as the checkpoint logfile, with
                      the string ".CKP" appended to it.  Therefore, this
                      id must conform to local file naming standards.

    void *save_area - A pointer to a programmer supplied data object in
                      memory.  This data object should correspond
                      exactly to the one used when the checkpoint()
                      function was called.

RETURN VALUE:
    -1 - The program was attempting to restart but could not.  This is
         usually due to a permission problem such as the inability to open
         an existing checkpoint logfile, or unable to create a checkpoint
         logfile. This is a serious error.
     0 - The program is starting normally.
     1 - The program is being restarted.

    SEE ALSO:
    ckp_close(), ckp_open(), checkpoint().
